#!/usr/bin/perl
#
# Copyright 2014-2015, Julian Catchen <jcatchen@illinois.edu>
#

use strict;

use constant false => 0;
use constant true  => 1;
use constant gtf   => 2;
use constant gff   => 3;

my $debug    = 0;
my $agp_path = "";
my $gtf_path = "";
my $unp_path = "";
my $out_path = "";
my $ann_type = gtf;

my (%agp, %agp_index, %unplaced_scaffolds);

parse_command_line();

load_agp_file($agp_path, \%agp);

load_unplaced_scaffolds($unp_path, \%unplaced_scaffolds);

index_agp(\%agp, \%agp_index);

parse_gtf($gtf_path, $out_path, \%agp, \%agp_index, \%unplaced_scaffolds);

sub parse_gtf {
    my ($gtf_path, $out_path, $agp, $index, $unplaced) = @_;

    my ($out_fh, $in_fh, @gene, $gtf_file, $scaf, $sbp, $ebp, $i, $found);

    if ($gtf_path =~ /.gz$/) {
        open($in_fh, "gunzip -c $gtf_path |") or die("Unable to open GTF input file: $gtf_path, $!\n");
    } else {
        open($in_fh, "<$gtf_path") or die("Unable to open GTF input file: $gtf_path, $!\n");
    }

    #
    # Pull the GTF filename off the path.
    #
    if ($gtf_path !~ /\//) {
	$gtf_file = $gtf_path;
    } else {
	($gtf_file) = ($gtf_path =~ /^.+\/(.+\.g[ft]f\.?g?z?)$/);
    }

    if ($ann_type == gtf) {
	$gtf_file =~ s/\.gtf\.?g?z?$/-translated.gtf/;
    } else {
	$gtf_file =~ s/\.gff\.?g?z?$/-translated.gff/;
    }
    open($out_fh, ">" . $out_path . "/" . $gtf_file) or die("Unable to open GTF output file: $out_path/$gtf_file, $!\n");

    my ($line, $gene_id, $next_gene_id, $parts, @components, $aref);
    my ($com_not_trans, $com_not_found, $com_trans_err, $com_success, %scaffolds, %lgs, %missing_scaffolds, @fails);

    my $mult_scaf = 0;
    my $not_found = 0;
    my $not_trans = 0;
    my $trans     = 0;
    my $error     = 0;
    my $total     = 0;
    my $gene_success   = 0;
    my $gene_fail      = 0;
    my $gene_not_found = 0;
    my $gene_mult_lgs  = 0;

    #
    # We need to read all the GTF components for a gene at one time. If all components
    # cannot be mapped, map none of them and report an error.
    #

    #
    # Get the first non-comment line.
    #
    $line       = "";
    @components = ();

    while ($line = <$in_fh>) {
	if ($line =~ /^\s*#/) {
	    print $out_fh $line;
	} else {
	    last;
	}
    }

    #
    # Determine the gene ID for this set of components.
    #
    chomp $line;
    $parts        = [];
    @{$parts}     = split(/\t/, $line);
    $gene_id      = find_gene_id($parts->[8]);
    $next_gene_id = $gene_id;
    push(@components, $parts);

    do {
	#
	# Retrieve the remaining components for this gene.
	#
	do {
            do {
                #
                # Skip lines with comments.
                #
                $line = <$in_fh>;
            } while ($line =~ /^\s*#/);

	    chomp $line;
	    $parts        = [];
	    @{$parts}     = split(/\t/, $line);
	    $next_gene_id = find_gene_id($parts->[8]);
	    push(@components, $parts) if ($next_gene_id eq $gene_id);
	} while ($next_gene_id eq $gene_id);

	$com_not_trans = 0;
	$com_not_found = 0;
	$com_trans_err = 0;
	$com_success   = 0;
	@fails         = ();
	%scaffolds     = ();
	%lgs           = ();

	#
	# Process all components together.
	#
	foreach $aref (@components) {
	    $scaf = $aref->[0];
	    $sbp  = $aref->[3];
	    $ebp  = $aref->[4];

	    #
	    # Test that all components for this gene are on the same scaffold prior to translation.
	    #
	    $scaffolds{$scaf}++;

	    #
	    # Find this scaffold in our assembly, translate the coordinates.
	    #
	    if (!defined($unplaced->{$scaf}) && !defined($index->{$scaf})) {
		$missing_scaffolds{$scaf}++;
		$com_not_found++;
		push(@fails, $aref->[2]);
		next;
	    } elsif (defined($unplaced->{$scaf})) {
		$com_not_trans++;
		next;
	    }

	    #
	    # Find the proper fragment of this scaffold that this component falls within.
	    #
	    $found = false;
	    for ($i = 0; $i < scalar(@{$index->{$scaf}}); $i++) {
		# #
		# # If this scaffold is in a negative orientation, reverse the basepair coordinates.
		# #
		# if ($index->{$scaf}->[$i]->{'ori'} eq "-") {
		# }

		if ($sbp >= $index->{$scaf}->[$i]->{'sbp'} &&
		    $sbp <  $index->{$scaf}->[$i]->{'ebp'} &&
		    $ebp >  $index->{$scaf}->[$i]->{'sbp'} &&
		    $ebp <= $index->{$scaf}->[$i]->{'ebp'}) {

		    $found     = true;
		    $aref->[0] = $index->{$scaf}->[$i]->{'chr'};
		    $aref->[3] = $index->{$scaf}->[$i]->{'chr_sbp'} + ($sbp - $index->{$scaf}->[$i]->{'sbp'});
		    $aref->[4] = $index->{$scaf}->[$i]->{'chr_sbp'} + ($ebp - $index->{$scaf}->[$i]->{'sbp'});
		    $com_success++;

		    #
		    # Test whether components span chromosomes.
		    #
		    $lgs{$index->{$scaf}->[$i]->{'chr'}}++;

		    # print STDERR "Translating $scaf to $index->{$scaf}->[$i]->{'chr'}; $sbp to $aref->[3]; $ebp to $aref->[4]\n";

		} elsif ($sbp >= $index->{$scaf}->[$i]->{'sbp'} &&
			 $sbp <  $index->{$scaf}->[$i]->{'ebp'} &&
			 $ebp >  $index->{$scaf}->[$i]->{'sbp'} &&
			 $ebp >  $index->{$scaf}->[$i]->{'ebp'}) {
		    $found  = true;
		    $com_trans_err++;
		    push(@fails, $aref->[2]);

		    # print STDERR 
		    # 	"Failed to translate: $scaf to $index->{$scaf}->[$i]->{'chr'}; ",
		    # 	"ann start: $sbp; chr start: $index->{$scaf}->[$i]->{'sbp'}; ",
		    # 	"ann end: $ebp; chr end: $index->{$scaf}->[$i]->{'ebp'}\n";

		} elsif ($sbp <  $index->{$scaf}->[$i]->{'sbp'} &&
			 $sbp <  $index->{$scaf}->[$i]->{'ebp'} &&
			 $ebp >  $index->{$scaf}->[$i]->{'sbp'} &&
			 $ebp <= $index->{$scaf}->[$i]->{'ebp'}) {
		    $found  = true;
		    $com_trans_err++;
		    push(@fails, $aref->[2]);

		    # print STDERR 
		    # 	"Failed to translate: $scaf to $index->{$scaf}->[$i]->{'chr'}; ",
		    # 	"ann start: $sbp; chr start: $index->{$scaf}->[$i]->{'sbp'}; ",
		    # 	"ann end: $ebp; chr end: $index->{$scaf}->[$i]->{'ebp'}\n";
		}

		last if ($found);
	    }

	    if ($found == false) {
		print STDERR "ERROR: Unable to find component ", join(",", @{$aref}), "\n";
	    }
	}

	$total     += scalar(@components);
	$not_trans += $com_not_trans;
	$not_found += $com_not_found;
	$error     += $com_trans_err;
	$trans     += $com_success;
	$mult_scaf++ if (scalar(keys %scaffolds) > 1);

	if ($com_success == scalar(@components)) {
	    #
	    # Output all the components;
	    #
	    foreach $aref (@components) {
		print $out_fh join("\t", @{$aref}), "\n";
	    }
	    $gene_success++;

	} elsif ($com_not_trans == scalar(@components)) {
	    #
	    # Output all the components;
	    #
	    foreach $aref (@components) {
		print $out_fh join("\t", @{$aref}), "\n";
	    }
	    $gene_success++;

	} elsif ($com_not_found == scalar(@components)) {
	    $gene_not_found++;

	} else {
	    print STDERR 
		"Unable to translate $gene_id; $com_success out of ", scalar(@components), 
		" components succeeded. Fails included: ", join(", ", @fails);
	    if (scalar(keys %lgs) > 1) {
		$gene_mult_lgs++;
		print STDERR "; spans multiple chrs: ", join(", ", keys %lgs);
	    }
	    print STDERR "\n";
	    $gene_fail++;
	}

	#
	# Prepare for the next component.
	#
	@components = ();
	push(@components, $parts);
	$gene_id = $next_gene_id;

    } while (!eof($in_fh));

    close($in_fh);
    close($out_fh);

    print STDERR 
	"Scaffolds in the GTF file not found in the assembly: ", join(", ", keys %missing_scaffolds), "\n",
	"$mult_scaf genes occurred on multiple scaffolds prior to translation.\n",
	"Processed ", $gene_success + $gene_fail, " genes; $gene_success translated, $gene_fail failures ($gene_mult_lgs on multiple chrs); ", 
	"$gene_not_found genes on mising scaffolds\n",
	"Processed $total components; ", 
	"$trans translations; $error translation errors; ", 
	"$not_trans components on unassembled scaffolds; ", 
	"$not_found components in ", scalar(keys %missing_scaffolds), " missing scaffolds.\n";
}

sub find_gene_id {
    my ($comment) = @_;

    my (@parts, $str, $gene_id);

    @parts = split(/;/, $comment);

    foreach $str (@parts) {
	if ($ann_type == gtf) {
	    if ($str =~ /^\s*gene_id ".+"\s*$/) {
		($gene_id) = ($str =~ /^\s*gene_id "(.+)"\s*$/);
		return $gene_id;
	    }
	} else {
	    if ($str =~ /^\s*ID=.+\s*$/) {
		($gene_id) = ($str =~ /^\s*ID="?(.+)"?\s*$/);
		return $gene_id;
	    }
	}
    }

    return undef;
}

sub index_agp {
    my ($agp, $index) = @_;

    my ($chr, $scaf, $href);

    foreach $chr (keys %{$agp}) {

	foreach $href (@{$agp->{$chr}}) {

	    next if ($href->{'type'} eq "N");

	    if (!defined($index->{$href->{'scaf'}})) {
		$index->{$href->{'scaf'}} = [];
	    }

	    push(@{$index->{$href->{'scaf'}}}, 
		 {'sbp'     => $href->{'scaf_sbp'},
		  'ebp'     => $href->{'scaf_ebp'},
		  'ori'     => $href->{'scaf_ori'},
		  'chr_sbp' => $href->{'start'},
		  'chr_ebp' => $href->{'end'},
		  'chr'     => $href->{'chr'}
		 });
	}
    }

    #
    # Sort all the scaffold components according to starting basepair.
    #
    foreach $scaf (keys %{$index}) {
	@{$index->{$scaf}} = sort {$a->{'sbp'} <=> $b->{'sbp'}} @{$index->{$scaf}};
    }

    # foreach $scaf (keys %{$index}) {
    # 	print STDERR $scaf, ":\n";
    # 	foreach $href (@{$index->{$scaf}}) {
    # 	    print STDERR "  ", $href->{'sbp'}, " - ", $href->{'ebp'}, "; Chr ", $href->{'chr'}, "\n";
    # 	}
    # }
}

sub load_agp_file {
    my ($agp_path, $agp) = @_;

    my ($fh, @parts, $line, $href);

    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    open($fh, "<" . $agp_path) or die("Unable to open AGP file '$agp_path' $!\n");

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	next if (substr($parts[0], 0, 1) eq "#");

	$href = {};
	$href->{'chr'}   = $parts[0];
	$href->{'start'} = $parts[1] + 0;
	$href->{'end'}   = $parts[2] + 0;
	$href->{'part'}  = $parts[3];
	$href->{'type'}  = $parts[4];
	$href->{'len'}   = $href->{'type'} eq "N" ? $parts[5] : $parts[7] - $parts[6] + 1;

	if ($href->{'type'} ne "N") {
	    $href->{'scaf'}     = $parts[5];
	    $href->{'scaf_sbp'} = $parts[6] + 0;
	    $href->{'scaf_ebp'} = $parts[7] + 0;
	    $href->{'scaf_ori'} = $parts[8];
	}

	push(@{$agp->{$href->{'chr'}}}, $href);
    }

    print STDERR "Loaded gap information for ", scalar(keys %{$agp}), " chromosomes.\n";

    close($fh);
}

sub load_unplaced_scaffolds {
    my ($unp_path, $unplaced) = @_;

    my ($fh, @parts, $line, $href);

    return if (length($unp_path) == 0);

    open($fh, "<" . $unp_path) or die("Unable to open unplaced scaffolds AGP file '$unp_path' $!\n");

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	next if (substr($parts[0], 0, 1) eq "#");

	$unplaced->{$parts[0]}++;
    }

    print STDERR "Loaded information for ", scalar(keys %{$unplaced}), " scaffolds.\n";

    close($fh);
}

sub parse_command_line {
    my $type;
    
    while (@ARGV) {
	$_ = shift @ARGV;
	if    ($_ =~ /^-g$/) { $gtf_path = shift @ARGV; }
	elsif ($_ =~ /^-a$/) { $agp_path = shift @ARGV; }
	elsif ($_ =~ /^-o$/) { $out_path = shift @ARGV; }
	elsif ($_ =~ /^-u$/) { $unp_path = shift @ARGV; }
	elsif ($_ =~ /^-t$/) { $type     = shift @ARGV; }
	elsif ($_ =~ /^-d$/) { $debug++; }
	elsif ($_ =~ /^-h$/) { usage(); }
	else {
	    print STDERR "Unknown command line option: '$_'\n";
	    usage();
	}
    }

    if (length($agp_path) == 0) {
	print STDERR "You must specify the AGP file defining assembly gaps.\n";
	usage();
    }

    if (length($gtf_path) == 0) {
	print STDERR "You must specify the GTF file to translate.\n";
	usage();
    }

    if (length($out_path) == 0 || !-e $out_path) {
	print STDERR "You must specify a directory to output translated GTF file.\n";
	usage();
    }
    $out_path .= "/" if (substr($out_path, -1) ne "/");

    if ($type eq "gff") {
	$ann_type = gff;
    } elsif ($type eq "gtf") {
	$ann_type = gtf;
    } else {
	print STDERR "Unknown file type, '$type'\n";
	usage();
    }
}

sub usage {
    print STDERR <<EOQ; 
translate_gtf.pl -a agp_path -g gtf_path -o out_path [-u agp_path] [-t type] [-d] [-h]
    a: path to the AGP formatted file defining assembly gaps.
    g: path to the GTF file to translate.
    u: path to the unplaced scaffolds AGP file.
    o: directory to output translated GTF file.
    t: input file type, either 'gtf' or 'gff'.
    h: display this help message.
    d: turn on debug output.

EOQ

exit(0);
}
