// -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
//
// Copyright 2016, Julian Catchen <jcatchen@illinois.edu>
//
// This file is part of Chromonomer.
//
// Chromonomer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Chromonomer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef __SCAFFOLD_H__
#define __SCAFFOLD_H__

#include <utility>
using std::pair;
using std::make_pair;
#include <string>
using std::string;
#include <map>
using std::map;
#include <set>
using std::set;
#include <vector>
using std::vector;

#include "constants.h"
#include "marker.h"

class OldScafComp {
public:
    string      scaffold;
    uint        start;
    uint        end;
    uint        part_num;

    OldScafComp() {
	this->start    = 0;
	this->end      = 0;
	this->part_num = 0;
    }
    OldScafComp(string scaffold, uint start, uint end) {
	this->scaffold = scaffold;
	this->start    = start;
	this->end      = end;
	this->part_num = 0;
    }
    OldScafComp(string scaffold, uint start, uint end, uint part_num) {
	this->scaffold = scaffold;
	this->start    = start;
	this->end      = end;
	this->part_num = part_num;
    }
};

class Scaffold {
public:
    string    id;
    uint      start_bp;
    uint      end_bp;
    uint      len;
    string    lg;
    ori_type  ori;

    //
    // Markers occuring on this scaffold.
    //
    vector<Marker> markers;
    //
    // Nodes in the map this scaffold occurs in.
    //
    set<double> nodes;

    //
    // An attempt to split this scaffold (at least partially) failed.
    //
    bool split_failure;

    vector<OldScafComp> old;

    Scaffold() {
	this->start_bp = 0;
	this->end_bp   = 0;
	this->len      = 0;
	this->ori      = ori_unknown;
	this->split_failure = false;
    }
    Scaffold(string id) {
	this->id       = id;
	this->start_bp = 0;
	this->end_bp   = 0;
	this->len      = 0;
	this->ori      = ori_unknown;
	this->split_failure = false;
    }
    Scaffold(string id, uint start, uint end, uint len, ori_type orientation) {
	this->id       = id;
	this->start_bp = start;
	this->end_bp   = end;
	this->len      = len;
	this->ori      = orientation;
	this->split_failure = false;
    }
    Scaffold(string id, uint start, uint end, uint len, ori_type orientation, bool failure) {
	this->id            = id;
	this->start_bp      = start;
	this->end_bp        = end;
	this->len           = len;
	this->ori           = orientation;
	this->split_failure = failure;
    }
};

class ScafAnn {
public:
    uint     agp_pos;
    uint     segment;
    ann_type type;

    ScafAnn(uint pos, uint seg, ann_type t) {
        this->agp_pos = pos;
        this->segment = seg;
        this->type    = t;
    }
    ScafAnn() {
        this->agp_pos = 0;
        this->segment = 0;
        this->type    = seg_beg;
    }
};

class ScafComp {
public:
    string      scaffold;
    comp_type   type;
    string      contig_id;
    uint        start;
    uint        end;
    uint        part_num;
    uint        len;
    uint        comp_beg;
    uint        comp_end;
    ori_type    ori;
    string      gap_type;
    bool        linkage;

    bool        split_fail;

    vector<ScafAnn>     annotations;
    vector<OldScafComp> old;

    ScafComp() {
	this->start      = 0;
	this->end        = 0;
	this->part_num   = 0;
	this->len        = 0;
	this->comp_beg   = 0;
	this->comp_end   = 0;
	this->ori        = ori_unknown;
	this->linkage    = false;
	this->split_fail = false;
    }
    ScafComp(string scaffold, comp_type type, uint start, uint end, uint part_num, uint len, 
	     string gap_type, bool linkage) {
	this->scaffold   = scaffold;
	this->type       = type;
	this->start      = start;
	this->end        = end;
	this->part_num   = part_num;
	this->len        = len;
	this->gap_type   = gap_type;
	this->linkage    = linkage;
	this->split_fail = false;
    }
    ScafComp(string scaffold, comp_type type, string contig_id, uint start, uint end, uint part_num, uint len, 
	     uint comp_beg, uint comp_end, ori_type ori) {
	this->scaffold   = scaffold;
	this->type       = type;
	this->start      = start;
	this->end        = end;
	this->part_num   = part_num;
	this->len        = len;
	this->comp_beg   = comp_beg;
	this->comp_end   = comp_end;
	this->ori        = ori;
	this->contig_id  = contig_id;
	this->split_fail = false;
    }

    int relocate(string scaffold, uint start, uint end) {
	//
	// Record the previous position.
	//
	this->old.push_back(OldScafComp(this->scaffold, this->start, this->end));

	this->scaffold     = scaffold;
	this->start        = start;
	this->end          = end;

	return 0;
    }
};

class ProbScaf {
public:
    string id;
    string lg;
    uint   splits;
    uint   conflict_markers;
    bool   split_fail;
    ProbScaf() {
	this->splits           = 0;
	this->conflict_markers = 0;
	this->split_fail       = false;
    }
    ProbScaf(string id) {
	this->id               = id;
	this->splits           = 0;
	this->conflict_markers = 0;
	this->split_fail       = false;
    }
};

#endif // __SCAFFOLD_H__
